@@ -4,8 +4,26 @@ import html5lib
 from html5lib import treebuilders
 from xml.dom import minidom, Node
 
+from htmltreediff.text import WordMatcher
+
 ## DOM utilities ##
 # parsing and cleaning #
+from xml.dom.pulldom import SAX2DOM
+import lxml.html, lxml.etree, lxml.sax
+def parse_lxml_dom(xml, html=True):
+    if html:
+        parse_func = lxml.html.document_fromstring
+    else:
+        parse_func = lxml.etree.fromstring
+    try:
+        tree = parse_func(xml)
+    except lxml.etree.XMLSyntaxError:
+        tree = parse_func('<body>%s</body>' % xml)
+
+    handler = SAX2DOM()
+    lxml.sax.saxify(tree, handler)
+    return handler.document
+
 def parse_minidom(xml, clean=True, html=True):
     html_parser = html5lib.HTMLParser(tree=treebuilders.getTreeBuilder('dom'))
     # Preprocessing
@@ -13,15 +31,14 @@ def parse_minidom(xml, clean=True, html=True):
     if clean:
         xml = remove_newlines(xml)
         xml = normalize_entities(xml)
+    xml = xml.strip()
+
     # Parse
-    if html:
-        dom = html_parser.parse(xml)
-        if clean:
-            remove_insignificant_text_nodes(dom)
-    else:
-        dom = minidom.parseString(xml)
+    dom = parse_lxml_dom(xml, html=html)
 
     if clean:
+        if html:
+            remove_insignificant_text_nodes(dom)
         # clean up irrelevant content
         for node in list(walk_dom(dom)):
             if node.nodeType == Node.COMMENT_NODE:
@@ -33,6 +50,18 @@ def parse_minidom(xml, clean=True, html=True):
             elif node.nodeName == 'span':
                 unwrap(node)
     dom.normalize()
+
+    # Make sure that the body element is the top of the dom.
+    for head_element in dom.getElementsByTagName('head'):
+        remove_node(head_element)
+    for html_element in dom.getElementsByTagName('html'):
+        unwrap(html_element)
+    if not dom.documentElement:
+        dom = parse_lxml_dom('', html=False)
+
+    if html:
+        assert dom.documentElement.tagName == 'body'
+
     return dom
 
 def remove_comments(xml):
@@ -40,8 +69,16 @@ def remove_comments(xml):
     Remove comments, as they can break the xml parser.
 
     See html5lib issue #122 ( http://code.google.com/p/html5lib/issues/detail?id=122 ).
+
+    >>> remove_comments('<!-- -->')
+    ''
+    >>> remove_comments('<!--\\n-->')
+    ''
+    >>> remove_comments('<p>stuff<!-- \\n -->stuff</p>')
+    '<p>stuffstuff</p>'
     """
-    return re.sub(r'<!--.*?-->', '', xml)
+    regex = re.compile(r'<!--.*?-->', re.DOTALL)
+    return regex.sub('', xml)
 
 def remove_newlines(xml):
     r"""Remove newlines in the xml.
@@ -71,6 +108,8 @@ def minidom_tostring(dom, pretty=False):
     else:
         xml = dom.toxml()
     xml = remove_xml_declaration(xml)
+    if xml == '<body/>':
+        return ''
     if xml.startswith('<body>') and xml.endswith('</body>'):
         xml = xml[len('<body>'):-len('</body>')]
     xml = dedent(xml.replace('\t', '  ')).strip()
@@ -93,6 +132,9 @@ class HashableNode(object):
                 self.node.nodeValue == other.node.nodeValue and
                 attribute_dict(self.node) == attribute_dict(other.node))
 
+    def __ne__(self, other):
+        return not self.__eq__(other)
+
     def __hash__(self):
         attributes = frozenset(attribute_dict(self.node).items())
         return hash((self.node.nodeType,
@@ -113,6 +155,33 @@ class HashableTree(object):
         child_hashes = hash(tuple(HashableTree(c) for c in self.node.childNodes))
         return hash((HashableNode(self.node), child_hashes))
 
+class FuzzyHashableTree(object):
+    cutoff = 0.4
+
+    def __init__(self, node):
+        self.node = node
+
+    def __eq__(self, other):
+        if HashableNode(self.node) != HashableNode(other.node):
+            return False
+
+        # Check for an exact tree match.
+        if HashableTree(self.node) == HashableTree(other.node):
+            return True
+
+        # Check for a fuzzy match.
+        ratio = tree_text_ratio(self.node, other.node)
+        if ratio >= self.cutoff:
+            return True
+
+        return False
+
+    def __hash__(self):
+        # This will never be equal if the top level tag in the tree is
+        # different. Beyond that, we can't make any guarantees.
+        return hash(HashableNode(self.node))
+
+
 def attribute_dict(node):
     if not node.attributes:
         return {}
@@ -123,9 +192,9 @@ def attribute_dict(node):
 
 def normalize_entities(html):
     # turn &nbsp; and aliases into normal spaces
-    html = html.replace('&nbsp;', ' ')
-    html = html.replace('&#160;', ' ')
-    html = html.replace('&#xA0;', ' ')
+    html = html.replace(u'&nbsp;', u' ')
+    html = html.replace(u'&#160;', u' ')
+    html = html.replace(u'&#xA0;', u' ')
     html = html.replace(u'\xa0', u' ')
     return html
 
@@ -193,18 +262,20 @@ def ancestors(node):
         yield ancestor
         ancestor = ancestor.parentNode
 
-def walk_dom(dom_or_node, elements_only=False):
+def walk_dom(dom, elements_only=False):
+    # allow calling this on a document as well as as node
+    if hasattr(dom, 'documentElement'):
+        dom = dom.documentElement
     def walk(node):
+        if not node:
+            return
         if elements_only and not is_element(node):
             return
         yield node
         for child in node.childNodes:
             for descendant in walk(child):
                 yield descendant
-    # allow calling this on a document as well as as node
-    if hasattr(dom_or_node, 'documentElement'):
-        return walk(dom_or_node.documentElement)
-    return walk(dom_or_node)
+    return walk(dom)
 
 def tree_text(node):
     """Return all the text below the given node as a single string.
@@ -218,6 +289,17 @@ def tree_text(node):
             text_list.append(descendant.nodeValue)
     return ' '.join(text_list)
 
+def tree_text_ratio(a_dom, b_dom):
+    """Compare two dom trees for text similarity, as a ratio."""
+    matcher = _text_matcher(a_dom, b_dom)
+    return matcher.text_ratio()
+
+def _text_matcher(a_dom, b_dom):
+    return WordMatcher(
+        a=tree_text(a_dom),
+        b=tree_text(b_dom),
+    )
+
 # manipulation #
 def copy_dom(dom):
     new_dom = minidom.Document()
@@ -244,27 +326,6 @@ def insert_or_append(parent, node, next_sibling):
     else:
         parent.appendChild(node)
 
-def remove_text_section(node, start, end):
-    """
-    Remove a section of text. Return the three parts created by the split.
-    The remaining adjacent text nodes are NOT normalized. The original text
-    node is removed from the document.
-    """
-    assert node.nodeType == Node.TEXT_NODE
-    text = node.nodeValue
-    parent = node.parentNode
-    def add_text_node(node_value):
-        text_node = node.ownerDocument.createTextNode(node_value)
-        if parent:
-            parent.insertBefore(text_node, node)
-        return text_node
-    before  = add_text_node(text[:start])
-    deleted = add_text_node(text[start:end])
-    after   = add_text_node(text[end:])
-    remove_node(deleted)
-    remove_node(node)
-    return before, deleted, after
-
 def wrap(node, tag):
     """Wrap the given tag around a node."""
     wrap_node = node.ownerDocument.createElement(tag)
